/****************************************************************************
**
**	QtTicTacToe  (C)  22 NOV 2002
**
**	David A.S.Crofts
**	23 Brisbane Street
**	Berwick Victoria 3806
**	Australia
**
**	email: dasc1961@netscape.net
**
*****************************************************************************/

#include "tictac.h"
#include <qapplication.h>
#include <qpainter.h>
#include <qdrawutil.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qlayout.h>
#include <stdlib.h>
#include <qdatetime.h>


// --------------------------------------------------------------------------
// TicTacState::copy_state
// --------------------------------------------------------------------------

void TicTacState::copy_state( TicTacState *q )
{
    int i;

    for (i = 0; i<9; ++i)
	q->board[i] = board[i];
    q->turn = turn;
    q->piece = piece;
}


// --------------------------------------------------------------------------
// TicTacState::gen_sucessors
// --------------------------------------------------------------------------

void *TicTacState::gen_sucessors()
{
    TicTacSucessor *q;
    TicTacState *New;
    unsigned int i;
    
    q = new TicTacSucessor;
    q->num = 0;

    if (value() == 0) {
    	for (i = 0; i<9; ++i) {
	    if (board[i] == Blank) {
	    	New = new TicTacState;
	    	copy_state( New);
	    	New->board[i] = turn;
	    	if (turn == Circle)
		    New->turn = Cross;
	    	else
		    New->turn = Circle;
	        q->ptr[q->num] = New;
	        ++(q->num);
	    }
	}
    }
    return(q);
}    


// --------------------------------------------------------------------------
// TicTacState::value
// --------------------------------------------------------------------------

int TicTacState::value()
{
    int x;
    
    x = 0;

    if (((board[0] == Cross) &&
	 (board[1] == Cross) &&
	 (board[2] == Cross)) ||
  	((board[3] == Cross) &&
	 (board[4] == Cross) &&
	 (board[5] == Cross)) ||
  	((board[6] == Cross) &&
	 (board[7] == Cross) &&
	 (board[8] == Cross)) ||

        ((board[0] == Cross) &&
	 (board[3] == Cross) &&
	 (board[6] == Cross)) ||
  	((board[1] == Cross) &&
	 (board[4] == Cross) &&
	 (board[7] == Cross)) ||
  	((board[2] == Cross) &&
	 (board[5] == Cross) &&
	 (board[8] == Cross)) ||

        ((board[0] == Cross) &&
	 (board[4] == Cross) &&
	 (board[8] == Cross)) ||
  	((board[2] == Cross) &&
	 (board[4] == Cross) &&
	 (board[6] == Cross)))

	x = 30000;

    else if

       (((board[0] == Circle) &&
	 (board[1] == Circle) &&
	 (board[2] == Circle)) ||
  	((board[3] == Circle) &&
	 (board[4] == Circle) &&
	 (board[5] == Circle)) ||
  	((board[6] == Circle) &&
	 (board[7] == Circle) &&
	 (board[8] == Circle)) ||

        ((board[0] == Circle) &&
       	 (board[3] == Circle) &&
	 (board[6] == Circle)) ||
  	((board[1] == Circle) &&
	 (board[4] == Circle) &&
	 (board[7] == Circle)) ||
  	((board[2] == Circle) &&
	 (board[5] == Circle) &&
	 (board[8] == Circle)) ||

        ((board[0] == Circle) &&
	 (board[4] == Circle) &&
	 (board[8] == Circle)) ||
  	((board[2] == Circle) &&
	 (board[4] == Circle) &&
	 (board[6] == Circle)))

	x = -30000;

    if (turn == Circle)
        x = -x;
    return(x);
}


// --------------------------------------------------------------------------
// TicTacState::f
// --------------------------------------------------------------------------

int TicTacState::f(int alpha, int beta)
{
    TicTacSucessor *q;
    int t;
    unsigned int i;
    
    q = static_cast<TicTacSucessor *>(gen_sucessors());
    if (q->num==0)
	alpha = value();
    for (i = 0; i<q->num && alpha<beta; ++i) {
	t = -(q->ptr[i]->f(-beta, -alpha));
	if (t>alpha)
	    alpha = t;
    }
    q->free();
    return(alpha);
}


// --------------------------------------------------------------------------
// TicTacState::play
// --------------------------------------------------------------------------

void *TicTacState::play()
{
    TicTacSucessor *q, *r;
    int max, t;
    unsigned int i, j;
    
    r = new TicTacSucessor;
    r->num = 0;
    q = static_cast<TicTacSucessor *>(gen_sucessors());
    max = -40000;
    for (i = 0; i<q->num; ++i) {
	t = -(q->ptr[i]->f(-40000, 40000));
	if (t==max) {
	    j = r->num;
	    r->ptr[j] = new TicTacState;
	    q->ptr[i]->copy_state( r->ptr[j]);
	    r->ptr[j]->piece = max;
	    ++(r->num);
	} else if (t>max) {
	    max = t;
	    for (j = 0; j<r->num; ++j)
	        delete r->ptr[j];
	    r->num = 1;
	    r->ptr[0] = new TicTacState;
	    q->ptr[i]->copy_state( r->ptr[0]);
	    r->ptr[0]->piece = max;
	}
    }
    q->free();
    return(r);
}




// --------------------------------------------------------------------------
// TicTacSucessor::free
// --------------------------------------------------------------------------

void TicTacSucessor::free()
{
    unsigned int i;

    for (i = 0; i<num; ++i)
		delete ptr[i];
    delete this;
}




// --------------------------------------------------------------------------
// TicTacButton::TicTacButton
// --------------------------------------------------------------------------

TicTacButton::TicTacButton( QWidget *parent ) : QPushButton( parent )
{
    t = Blank;
}


// --------------------------------------------------------------------------
// TicTacButton::drawButtonLabel
// --------------------------------------------------------------------------

void TicTacButton::drawButtonLabel( QPainter *p )
{
    QColor border( colorGroup().foreground() );
    QColor fill( colorGroup().light() );

    int pad;
    QRect r = rect();
    
    pad = (r.width()+r.height())/16;
    
    p->setPen( QPen( border, (r.width()+r.height())/14 ) );

    if ( t == Circle ) {
	p->drawEllipse( 
	    r.left()+pad, 
	    r.top()+pad,
	    r.width()-2*pad, 
	    r.height()-2*pad );
    } else if ( t == Cross ) {			// draw cross
	p->drawLine( 
	    r.topLeft()+QPoint(pad,pad), 
	    r.bottomRight()-QPoint(pad,pad));
	p->drawLine( 
	    r.bottomLeft()+QPoint(pad,-pad),
	    r.topRight()-QPoint(pad,-pad));
    }
    p->setPen( QPen( fill, (r.width()+r.height())/14 -2 ) );
    if ( t == Circle ) {
	p->drawEllipse( 
	    r.left()+pad, 
	    r.top()+pad, 
	    r.width()-2*pad, 
	    r.height()-2*pad );
    } else if ( t == Cross ) {			// draw cross
	p->drawLine( 
	    r.topLeft()+QPoint(pad+1,pad+1),
	    r.bottomRight()-QPoint(pad+1,pad+1));
	p->drawLine( 
	    r.bottomLeft()+QPoint(pad+1,-(pad+1)),
	    r.topRight()-QPoint(pad+1,-(pad+1)));
    }
}





// --------------------------------------------------------------------------
// TicTacGameBoard::TicTacGameBoard
// --------------------------------------------------------------------------

TicTacGameBoard::TicTacGameBoard( QWidget *parent, const char *name )
    : QWidget( parent, name )
{
    st = Init;					// initial state
    comp_starts = FALSE;			// human starts
    buttons = new TicTacButtons(9);		// create real buttons
    btArray = new TicTacArray(9);		// create button model
    QGridLayout * grid = new QGridLayout( this, 3, 3, 4 );
    for ( int i=0; i<9; i++ ) {			// create and connect buttons
	TicTacButton *ttb = new TicTacButton( this );
	ttb->setEnabled( FALSE );
	connect( ttb, SIGNAL(clicked()), SLOT(buttonClicked()) );
	grid->addWidget( ttb, i%3, i/3 );
	buttons->insert( i, ttb );
	btArray->at(i) = Blank;	// initial button type
    }
    QTime t = QTime::currentTime();		// set random seed
    srand( t.hour()*12+t.minute()*60+t.second()*60 );
    
    TicTacGameBoard::newGame();
}


// --------------------------------------------------------------------------
// TicTacGameBoard::~TicTacGameBoard
// --------------------------------------------------------------------------

TicTacGameBoard::~TicTacGameBoard()
{
    delete buttons;
    delete btArray;
}


// --------------------------------------------------------------------------
// TicTacGameBoard::newGame
// --------------------------------------------------------------------------

void TicTacGameBoard::newGame()
{
    st = HumansTurn;
    for ( int i=0; i<9; i++ )
	btArray->at(i) = Blank;
    if ( comp_starts )
	computerMove();
    else
	updateButtons();
}


// --------------------------------------------------------------------------
// TicTacGameBoard::buttonClicked
// --------------------------------------------------------------------------

void TicTacGameBoard::buttonClicked()
{
   if ( st != HumansTurn )			// not ready
	return;
    int i = buttons->findRef( (TicTacButton*)sender() );
    TicTacButton *b = buttons->at(i);		// get piece that was pressed
    if ( b->type() == Blank ) {	// empty piece?
	btArray->at(i) = Cross;
	updateButtons();
        computerMove();
    }
}


// --------------------------------------------------------------------------
// TicTacGameBoard::updateButtons
// --------------------------------------------------------------------------

void TicTacGameBoard::updateButtons()
{
    for ( int i=0; i<9; i++ ) {
	if ( buttons->at(i)->type() != btArray->at(i) )
	    buttons->at(i)->setType( (ButtonType)btArray->at(i) );
	buttons->at(i)->setEnabled( 
	    buttons->at(i)->type() == Blank );
    }
}


// --------------------------------------------------------------------------
// TicTacGameBoard::computerMove
// --------------------------------------------------------------------------

void TicTacGameBoard::computerMove()
{

    int 		i;
    TicTacState	move;
    TicTacSucessor   *q;
    
    move.turn = Circle;
    for ( i=0; i<9; i++ )
	move.board[i] =  btArray->at(i);
    
    q = static_cast<TicTacSucessor *>(move.play());

    if ( q->num > 0)
        (q->ptr[rand()%q->num])->copy_state( &move );
    else if (q->num == 0 ) {
        st = NobodyWon;
	emit finished();
   } else 
        st = HumansTurn;

    q->free();

    if ( move.value() ) {
        st = ComputerWon;
	emit finished();
    } else {	
        q = static_cast<TicTacSucessor *>(move.gen_sucessors());
        if ( q->num == 0 ) {
            st = NobodyWon;
	    emit finished();
	}
    }
    for ( i=0; i<9; i++ )
	btArray->at(i) = move.board[i];

    updateButtons();				// update buttons
}





// --------------------------------------------------------------------------
// TicTacToe::TicTacToe
// --------------------------------------------------------------------------

TicTacToe::TicTacToe( QWidget *parent, const char *name )
    : QWidget( parent, name )
{
    QVBoxLayout * l = new QVBoxLayout( this, 6 );

    // Create a message label

    message = new QLabel( this );
    message->setFrameStyle( QFrame::Panel | QFrame::Sunken );
    message->setAlignment( AlignCenter );
    message->setFixedHeight( message->sizeHint().height() );
    l->addWidget( message );

    // Create the game board and connect the signal finished() to this
    // gameOver() slot

    board = new TicTacGameBoard( this );
    connect( board, SIGNAL(finished()), SLOT(gameOver()) );
    l->addWidget( board );

    // Create a horizontal frame line

    QFrame *line = new QFrame( this );
    line->setFrameStyle( QFrame::HLine | QFrame::Sunken );
    l->addWidget( line );

    // Create the push buttons and connect their clicked() signals
    // to this right slots.

    newGame = new QPushButton( "&New Game", this );
    connect( newGame, SIGNAL(clicked()), SLOT(newGameClicked()) );
    quit = new QPushButton( "&Quit", this );
    connect( quit, SIGNAL(clicked()), qApp, SLOT(quit()) );
    QHBoxLayout * b = new QHBoxLayout;
    l->addLayout( b );
    b->addWidget( newGame );
    b->addWidget( quit );

    newState();
}


// --------------------------------------------------------------------------
// TicTacToe::newGameClicked
// --------------------------------------------------------------------------

void TicTacToe::newGameClicked()
{
    if ( board->comp_starts ) {
	board->comp_starts = FALSE;
    } else
        board->comp_starts = TRUE;
    board->newGame();
    newState();
}


// --------------------------------------------------------------------------
// TicTacToe::gameOver
// --------------------------------------------------------------------------

void TicTacToe::gameOver()
{
    newState();					// update text box
}


// --------------------------------------------------------------------------
// TicTacToe::newState
// --------------------------------------------------------------------------

void TicTacToe::newState()
{
    static const char *msg[] = {		// TicTacGameBoard::State texts
	"Click Play to start", 
	"Make your move",
	"You Won !!!", 
	"I Win !!!", 
	"It's a draw !!!" };
    message->setText( msg[board->state()] );
    return;
}


