/****************************************************************************
**
**	QtTicTacToe  (C)  22 NOV 2002
**
**	David A.S.Crofts
**	23 Brisbane Street
**	Berwick Victoria 3806
**	Australia
**
**	email: dasc1961@netscape.net
**
*****************************************************************************/


#ifndef TICTAC_H
#define TICTAC_H


#include <qpushbutton.h>
#include <qvector.h>

class QComboBox;
class QLabel;


enum ButtonType { Blank, Cross, Circle };


// --------------------------------------------------------------------------
// TicTacState implements the game state.
// --------------------------------------------------------------------------

class TicTacState
{
public:
    void 	    *gen_sucessors();
    void	    *play();
    int 	    f( int, int );
    void 	    copy_state( TicTacState * );
    int		    value();
    ButtonType	    board[10];
    ButtonType	    turn;
    int		    piece;
};


// --------------------------------------------------------------------------
// TicTacSucessor implements the game sucessors.
// --------------------------------------------------------------------------

class TicTacSucessor
{
public:
    unsigned int    num;
    TicTacState  *ptr[10];
    void 	    free();
};


// --------------------------------------------------------------------------
// TicTacButton implements a single tic-tac-toe button
// --------------------------------------------------------------------------

class TicTacButton : public QPushButton
{
    Q_OBJECT
public:
    TicTacButton( QWidget *parent );
    
    ButtonType	type() const		
    { return t; }
    
    void	setType( ButtonType type )	
    { t = type; repaint(); }
    
    QSizePolicy sizePolicy() const
    { return QSizePolicy( QSizePolicy::Preferred, QSizePolicy::Preferred ); }
    
    QSize sizeHint() const 
    { return QSize( 75, 75 ); }
    
    QSize minimumSizeHint() const 
    { return QSize( 41, 41 ); }

protected:
    void	drawButtonLabel( QPainter * );

private:
    ButtonType t;
};


// Using template vector to make vector-class of TicTacButton.
// This vector is used by the TicTacGameBoard class defined below.

typedef QVector<TicTacButton>	TicTacButtons;
typedef QArray<ButtonType>	TicTacArray;


// --------------------------------------------------------------------------
// TicTacGameBoard implements the tic-tac-toe game board.
// TicTacGameBoard is a composite widget that contains N x N TicTacButtons.
// N is specified in the constructor.
// --------------------------------------------------------------------------

class TicTacGameBoard : public QWidget
{
    Q_OBJECT
public:
    TicTacGameBoard( QWidget *parent=0, const char *name=0 );
   ~TicTacGameBoard();
    enum	State { Init, HumansTurn, HumanWon, ComputerWon, NobodyWon };

    State	state() const
    { return st; }
    
    void        newGame();
    bool	comp_starts;

signals:
    void	finished();			// game finished

private slots:
    void	buttonClicked();

private:
    void        setState( State state )
    { st = state; }
    
    void	updateButtons();
    void 	computerMove();
    State	st;
    int		nBoard;
    TicTacArray *btArray;
    TicTacButtons *buttons;
};


// --------------------------------------------------------------------------
// TicTacToe implements the complete game.
// TicTacToe is a composite widget that contains a TicTacGameBoard and
// two push buttons for starting the game and quitting.
// --------------------------------------------------------------------------

class TicTacToe : public QWidget
{
    Q_OBJECT
public:
    TicTacToe( QWidget *parent=0, const char *name=0 );

private slots:
    void	newGameClicked();
    void	gameOver();

private:
    void	newState();
    QPushButton *newGame;
    QPushButton *quit;
    QLabel	*message;
    TicTacGameBoard *board;
};


#endif // TICTAC_H

